import requests
import os
import json
from datetime import datetime, date, timedelta
from kafka import KafkaProducer
from kafka.partitioner import RoundRobinPartitioner
from google.cloud import secretmanager
from pandas.api.types import is_numeric_dtype
import pendulum
import time
import numpy as np
import pandas as pd
import psycopg2
import pymysql
import logging

logger = logging.getLogger()
client = secretmanager.SecretManagerServiceClient()

try:
    secret = os.environ.get('prometheus_bootstrap')
    payload = json.loads(
        client.access_secret_version(secret).payload.data.decode('utf-8')
    )
    prometheus_bootstrap = payload.get('bootstrap')
except:
    raise Exception('please export a correct variable `prometheus_bootstrap`.')

try:
    secret = os.environ.get('kafka_bootstrap')
    payload = json.loads(
        client.access_secret_version(secret).payload.data.decode('utf-8')
    )
    kafka_bootstrap = payload.get('bootstrap')
except:
    raise Exception('please export a correct variable `kafka_bootstrap`.')

kafka_producer = KafkaProducer(
    bootstrap_servers = kafka_bootstrap.split(','),
    api_version = (0, 10),
    partitioner = RoundRobinPartitioner(),
)


def json_loads(row):
    """
    put uuid from waterhealer as `uuid` key in row.

    Parameters
    ----------
    row : (uuid, data)

    Returns
    -------
    result : Dict
    """
    load = json.loads(row[1])
    load['uuid'] = row[0]
    return load


# ----------------------------------------------------------------------
client = secretmanager.SecretManagerServiceClient()
secret = 'projects/fave-data/secrets/rds-production-replica/versions/1'
payload = json.loads(
    client.access_secret_version(secret).payload.data.decode('utf-8')
)
postgres_user = payload.get('user')
postgres_password = payload.get('password')
postgres_database = payload.get('database')

secret = 'projects/fave-data/secrets/rds-mysql-favefood/versions/1'
payload = json.loads(
    client.access_secret_version(secret).payload.data.decode('utf-8')
)
mysql_user = payload.get('user')
mysql_password = payload.get('password')
mysql_database = payload.get('database')

secret = 'projects/fave-data/secrets/cloudsql-warehouse/versions/1'
payload = json.loads(
    client.access_secret_version(secret).payload.data.decode('utf-8')
)
gcp_postgres_user = payload.get('user')
gcp_postgres_password = payload.get('password')
gcp_postgres_database = payload.get('database')

# ----------------------------------------------------------------------
# Destination DB for streaming/Source for bi_dwh.ref_cities
def get_gcp_replica():
    connection = psycopg2.connect(
        user = gcp_postgres_user,
        password = gcp_postgres_password,
        host = '10.46.192.5',
        port = 5432,
        database = gcp_postgres_database,
    )
    return connection


# ----------------------------------------------------------------------
# to merge with other tables
def get_rds_replica():
    connection = psycopg2.connect(
        user = postgres_user,
        password = postgres_password,
        host = '10.148.15.219',
        port = 5434,
        database = postgres_database,
    )
    return connection


# ----------------------------------------------------------------------
# to get food ordering
def get_mysql_replica():
    connection = pymysql.connect(
        user = mysql_user,
        password = mysql_password,
        host = '10.148.15.219',
        port = 3306,
        database = mysql_database,
    )
    return connection


# ----------------------------------------------------------------------
def millisecond_to_timestamp(ms, tz):

    if isinstance(ms, str):
        ms = int(ms)
    ##stored datetime values are in microseconds (10^-6). fromtimestamp f(x) expects seconds as an argument.
    ms_dt = datetime.fromtimestamp(ms / 1_000_000.0)

    if type(tz) == str:
        local_tz = pendulum.timezone(
            tz
        )  ##To convert the string timezone value to an actual timezone
        ##Localising timezone:
        ms_dt_local = datetime.fromtimestamp(ms / 1_000_000.0, tz = local_tz)
    else:
        ##When timezone not available, just copy unlocalised datetime into local dt variable
        ms_dt_local = ms_dt

    return (
        ##Epoch conversion to readable timestamp (output is a string)
        ms_dt.strftime('%Y-%m-%d %H:%M:%S.%f'),  ##unlocalised
        ms_dt_local.strftime('%Y-%m-%d %H:%M:%S.%f'),  ##localised
    )


# ----------------------------------------------------------------------
def get_main_category(parent_category_name):

    bnw = [
        'spafit',
        'beautyfit',
        'beauty',
        'beauty & wellness',
        'massage',
        'pamper',
    ]
    fnb = ['food & beverages', 'eat', 'ramadan', 'ramadhan']
    services = ['services', 'health & education']
    activities = ['hobby & leisure', 'activity', 'entertainment']
    travel = ['travel', 'staycation']
    kids = ['kids']
    others = [
        'fashion & jewelry',
        'home appliance',
        'gadget, tech & oto',
        'charity',
        'general',
    ]
    fitness = ['fitness']
    retail = ['retail', 'e_shop']

    if not parent_category_name:
        return 'Others'

    if parent_category_name.lower() in bnw:
        main_cat = 'B&W'
    elif parent_category_name.lower() in fnb:
        main_cat = 'F&B'
    elif parent_category_name.lower() in services:
        main_cat = 'Services'
    elif parent_category_name.lower() in activities:
        main_cat = 'Activities'
    elif parent_category_name.lower() in travel:
        main_cat = 'Travel'
    elif parent_category_name.lower() in kids:
        main_cat = 'Kids'
    elif parent_category_name.lower() in others:
        main_cat = 'Others'
    elif parent_category_name.lower() in fitness:
        main_cat = 'Fitness'
    elif parent_category_name.lower() in retail:
        main_cat = 'Retail'
    else:
        main_cat = 'Others'

    return main_cat


# ----------------------------------------------------------------------
def get_platform(user_agent):

    apps = ['global', 'kfit/', 'kfit-korea']

    user_agent = user_agent.lower()

    if any(app in user_agent for app in apps) is True:
        if 'android' in user_agent:
            platform = 'Android'
        elif 'ios' in user_agent:
            platform = 'iOS'
        else:
            platform = 'Other App'
    else:
        if 'mobile' in user_agent:
            platform = 'Mobile Web'
        elif 'android' in user_agent:
            platform = 'Mobile Web'
        elif 'ios' in user_agent:
            platform = 'Mobile Web'
        elif 'pwa' in user_agent:
            platform = 'PWA'
        elif user_agent == None or user_agent == '':
            platform = 'NA'
        else:
            platform = 'Desktop Web'

    return platform


# ----------------------------------------------------------------------
def get_payment_type(payment_type):

    if not payment_type:
        return 'NA'

    if payment_type == 0:  # input is integer
        payment_type = 'BrainTree'  # output is string
    elif payment_type == 1:
        payment_type = 'Adyen'
    elif payment_type == 2:
        payment_type = 'Veritrans'
    elif payment_type == 3:
        payment_type = 'Reseller'
    elif payment_type == 4:
        payment_type = 'OVO'
    elif payment_type == 5:
        payment_type = 'Airasia Big'
    elif payment_type == 6:
        payment_type = 'Boost'
    elif payment_type == 7:
        payment_type = 'GrabPay'
    elif payment_type == 8:
        payment_type = 'FLP Client'
    else:
        payment_type = 'NA'

    return payment_type


# ----------------------------------------------------------------------
def clean_local_price(price, schema_name):
    if not pd.isna(price):
        if schema_name == 'jakarta':
            # artificially inflated by *100.0 so don't have to selectively apply cents_converter
            output = (
                float(re.sub('[a-zA-Z]|\.|\,|\$|₱|' '|"', '', price)) * 100.0
            )
        else:
            if ',' in price:
                if (
                    len(price.split(',')[1]) >= 1
                    and len(price.split(',')[1]) <= 2
                ):
                    first_part = re.sub('.', ',', price.split(',')[0])
                    sec_part = re.sub(',', '.', price.split(',')[1])
                    comb = re.sub(
                        '[a-zA-Z]|\.|\,|\$|₱|' '|"',
                        '',
                        first_part + '.' + sec_part,
                    )
                    output = float(comb)
                else:
                    output = float(
                        re.sub('[a-zA-Z]|\.|\,|\$|₱|' '|"', '', price)
                    )
            else:
                output = float(re.sub('[a-zA-Z]|\.|\,|\$|₱|' '|"', '', price))

    else:
        output = None

    return output


# ----------------------------------------------------------------------
def cents_converter(table):
    ##id columns that need to be changed from cents
    column_names = list(table.columns)

    for i in range(len(table)):
        ##List of column_names
        cents_fields = [
            column_name
            for column_name in column_names
            if '_cent' in column_name
        ]

        for column_name in cents_fields:
            if column_name[-6:] == '_cents':
                new_column_name = column_name[:-6]
            elif column_name[-5:] == '_cent':
                new_column_name = column_name[:-5]
            ##Just in case cents_fields picks up field that is not meant to be /100.0
            else:
                new_column_name = column_name

            table[column_name] = table[column_name].replace({pd.np.nan: 0})
            table[column_name] = table[column_name].astype(float)
            table[new_column_name] = table[column_name] / 100.0

    return table


# ----------------------------------------------------------------------
##Fill financial figure's NA with 0
def zerorise_finance_fig(
    table, zerorise_list = None, zerorise_exclude_list = ['id']
):
    for i in list(table.columns):
        if i not in zerorise_exclude_list and ('_id' not in i and i != 'id'):
            if is_numeric_dtype(table[i]) or i in zerorise_list:
                table[i] = table[i].replace({pd.np.nan: 0})
    return table


# ----------------------------------------------------------------------
def cast_int(row):
    try:
        return int(row)
    except:
        return row


def integer_converter(table):
    for column in table.columns:
        if '_id' in column or column == 'id':
            if any([c in column for c in ['aabig']]):
                continue
            try:
                if object == table[column].dtype:
                    table[column] = table[column].apply(cast_int)
                else:
                    table[column] = table[column].astype('Int32')
                    table[column] = table[column].astype('int')
            except:
                pass
    return table


# ----------------------------------------------------------------------
def get_schema_details(table):

    ##Get unique schema_name
    schema_name = table['schema_name'].unique().tolist()
    if len(schema_name) == 1:
        ##Convert to string. Format needs to be so that it can be inputted in sql
        schema_name_str = str(schema_name)

    schema_name_str = (
        schema_name_str.replace('[', '')
        .replace(']', '')
        .replace("'", '')
        .replace(',', '')
    )

    ##Get unique currency
    # For reservation_sets table only, currency is nested in transaction_details
    if table['table'].unique()[0] == 'reservation_sets':

        ##None can be one of the elements so need to filter it out
        currency = list(filter(None, table['dtl_currency'].unique().tolist()))

    elif 'currency' in table.columns:
        currency = table['currency'].unique().tolist()

    # If table is not reservation_sets and does not have a currency column
    else:
        currency = ['Null']

    if len(currency) == 1:
        ##Format needs to be string to be inputted in SQL
        currency_str = str(currency)

        currency_str = (
            currency_str.replace('[', '')
            .replace(']', '')
            .replace("'", '')
            .replace(',', '')
        )

    schema_currency = [schema_name_str, currency_str]

    return schema_currency


# ----------------------------------------------------------------------
def tuple_to_str(list):
    tupled = tuple(list)
    tupled_str = str(tupled)
    if len(tupled) == 1:
        tupled_str = tupled_str.replace(',', '')
    if len(tupled) == 0:
        tupled_str = '(null)'
    return tupled_str


# ----------------------------------------------------------------------
def publish_kafka(row, topic):
    """
    Push a data to a kafka topic.
    Parameters
    ----------
    row : Dict
    topic : str
    Returns
    -------
    result : bool
    """
    try:
        value = json.dumps(row)
        value_bytes = bytes(value, encoding = 'utf-8')
        x = kafka_producer.send(topic, value = value_bytes)
        return True
    except Exception as ex:
        print('Exception in publishing message')
        print(str(ex))
        return False


# ----------------------------------------------------------------------
def push_to_kafka(row, topic):
    if row['payment_state'] in ['Successful', 'Cancelled/Refunded']:
        publish_kafka(row, topic)

    return row


# ----------------------------------------------------------------------
def separate_uuid(row):
    """
    separate uuid key from a dictionary.
    Parameters
    ----------
    row : Dict
    Returns
    -------
    result : (uuid, row)
    """

    if not row:
        return (None, None)

    uuid = row.pop('uuid')

    return (uuid, row)


# ----------------------------------------------------------------------
def push_prometheus(row, name, logging: bool = True):
    """
    push to prometheus HTTP.
    Parameters
    ----------
    row : Dict
    name : str
        name of prometheus metric.
    logging : bool, optional (default=True)
        print value in row before exit.
    Returns
    -------
    result : bool
    """
    success = False
    if row['data']['status']:
        try:
            requests.get(f'{prometheus_bootstrap}{name}', timeout = 3)
            success = True
        except Exception as e:
            if logging:
                print(e)
    if logging:
        print(datetime.now(), row)
    return success


# ----------------------------------------------------------------------
def groupby_sort(rows):

    if not len(rows):
        return []

    df = pd.DataFrame(rows)
    return df.sort_values('updated_at').to_dict('records')


# ----------------------------------------------------------------------
def remove_elements(rows, rejected):
    return [row for row in rows if row not in rejected]


# ----------------------------------------------------------------------
def push_elements(rows, schema, function, topic):
    for row in rows:
        d = {'id': row, 'schema': schema, 'function': function}
        print(d)
        publish_kafka(d, topic)


# ----------------------------------------------------------------------
def gather_uuid(rows, uuids):
    for i in range(len(rows)):
        uuids.append((rows['uuid'].iloc[i], {'status': True}))


# ----------------------------------------------------------------------
def loop_query(
    sql,
    conn,  ##connection
    stream_df,  ##streamed table
    match_on,  ##Column names to be used in matching (format: [streamed_df column_name, queried_table column_name])
    id_list,  ##list of ids from streamed df
    current_function,  ##state whether insert/update stream
    schema_str,  ##from get_schema_details
    loop_name,  ##state which loop this is (format: credits_transactions/external_trx_id/finance_journals)
    rejected_uuid,
    fail_kafka_topic,
    wait_period = 30,  ##seconds between each loop
    max_retry = 3,  ##number of loops to attempt
):

    ##Query using SQL and connection
    df = pd.read_sql(sql, conn)

    # Stop early if no success_ids
    if len(id_list) == 1 and id_list[0] == None:
        return df

    ##Count queried table's rows
    rows_len = len(df[match_on[1]].unique())

    ##initiate retry count
    retry = 0

    ##When number of ids don't match ...
    while len(id_list) != rows_len:

        ## ... and retries have been exhausted:
        if retry > max_retry:

            # Isolate ids that are present in stream but not in queried results
            not_in = set(id_list) - set(df[match_on[1]].unique().tolist())

            # Define message for logging
            message = {
                'function': current_function,
                'loop': f'{loop_name} loop',
                'e_card_purchases': str(id_list),
                'query_rows': str(df[match_on[1]].unique().tolist()),
                'not_in': str(not_in),
                'schema': schema_str,
            }
            logger.error(
                f'{schema_str} {current_function}, {loop_name} loop stuck',
                extra = message,
            )

            # For rows in stream that failed to match with queried results, append their uuids to the rejected_uuid list
            gather_uuid(
                stream_df.loc[stream_df[match_on[0]].isin(not_in)],
                rejected_uuid,
            )

            # Return only matchable rows from the streamed df
            stream_df = stream_df.loc[~stream_df[match_on[0]].isin(not_in)]

            # Return only matchable rows from the queried df
            df = df.loc[~df[match_on[1]].isin(not_in)]

            # Remove unmatchable ids from the id_list
            remove_elements(id_list, not_in)

            # Push unmatched elements to the failed uuid topic
            push_elements(
                not_in,
                schema = schema_str,
                function = f'{current_function}_{loop_name}_loop',
                topic = fail_kafka_topic,
            )

            break

        ## ... and retries have NOT been exhausted:
        # Wait for the given number of seconds
        time.sleep(wait_period)

        # Then query and count rows again and store in SAME variable. It will then loop to compare again.
        df = pd.read_sql(sql, conn)
        rows_len = len(df[match_on[1]].unique())
        # Add 1 attempt made to retry count
        retry += 1

    return df


# ----------------------------------------------------------------------
# Used in failed streaming script only
def update_failed_uuid(rows):

    if not len(rows):
        return []

    return rows[1]
